__author__ = 'labx'

import os, sys, numpy

from AnyQt.QtGui import QImage, QPixmap,  QTextCursor
from AnyQt.QtWidgets import QLabel, QWidget, QHBoxLayout, QMessageBox

from silx.gui.plot.ImageView import ImageView

import orangecanvas.resources as resources
from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from oasys2.widget.widget import OWAction
from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.util.widget_objects import TriggerIn, OasysThicknessErrorsData
from oasys2.widget.util.widget_util import EmittingStream
from oasys2.widget.gui import Styles
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from orangecontrib.shadow4.widgets.gui.ow_automatic_element import AutomaticElement
from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, ShadowPlot
from orangecontrib.shadow4.util.python_script import PythonScript
from orangecontrib.shadow4.util.shadow4_util import TriggerToolsDecorator

from hybrid_methods.coherence.hybrid_screen import HybridListener, \
    HybridCalculationType, HybridPropagationType, HybridDiffractionPlane, \
    HybridInputParameters, HybridCalculationResult, HybridGeometryAnalysis

from shadow4_advanced.hybrid.s4_hybrid_screen import S4HybridOE, S4HybridBeam, S4HybridScreen, S4HybridScreenElement

class HybridScreen(AutomaticElement, TriggerToolsDecorator, HybridListener):

    class Inputs:
        shadow_data          = Input("Shadow Data", ShadowData, default=True, auto_summary=False)
        trigger              = TriggerToolsDecorator.get_trigger_input()
        thickness_error_data = Input("Thickness Errors Data", OasysThicknessErrorsData, default=True, auto_summary=False)

    class Outputs:
        shadow_data_ff = Output("Output Data (Far Field)", ShadowData, default=True, auto_summary=False)
        shadow_data_nf = Output("Output Data (Near Field)", ShadowData, default=True, auto_summary=False)
        trigger        = TriggerToolsDecorator.get_trigger_output()

    name = "Hybrid Screen"
    description = "Shadow HYBRID: Hybrid Screen"
    icon = "icons/hybrid_screen.png"
    maintainer = "Luca Rebuffi and Xianbo Shi"
    maintainer_email = "lrebuffi(@at@)anl.gov, xshi(@at@)aps.anl.gov"
    priority = 1.0
    category = "HYBRID"
    keywords = ["data", "file", "load", "read"]

    want_control_area = 1
    want_main_area = 1

    view_type = Setting(1)
    send_original_beam = Setting(0)

    diffraction_plane = Setting(1)
    calculation_type = Setting(2)

    propagation_type = Setting(0)

    n_bins_x = Setting(100)
    n_bins_z = Setting(100)
    n_peaks = Setting(10)
    fft_n_pts = Setting(1e5)

    analyze_geometry = Setting(1)

    crl_error_profiles = Setting([])
    crl_material_data = Setting(0)
    crl_material = Setting("Be")
    crl_delta = Setting(1e-6)
    crl_scaling_factor = Setting(1.0)
    crl_coords_to_m = Setting(1.0)
    crl_thickness_to_m = Setting(1.0)

    __input_data = None
    __plotted_data = None

    TABS_AREA_HEIGHT = 560
    CONTROL_AREA_WIDTH = 405

    IMAGE_WIDTH = 860
    IMAGE_HEIGHT = 545

    def __init__(self):
        super().__init__()

        self.runaction = OWAction("Run Hybrid", self)
        self.runaction.triggered.connect(self.run_hybrid)
        self.addAction(self.runaction)

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal", width=self.CONTROL_AREA_WIDTH-5)

        button = gui.button(button_box, self, "Run HYBRID", callback=self.run_hybrid)
        button.setStyleSheet(Styles.button_blue)

        main_tabs = oasysgui.tabWidget(self.mainArea)
        plot_tab = oasysgui.createTabPage(main_tabs, "Plots")
        out_tab = oasysgui.createTabPage(main_tabs, "Output")
        script_tab = oasysgui.createTabPage(main_tabs, "Script")

        self.shadow4_script = PythonScript()
        self.shadow4_script.code_area.setFixedHeight(400)

        script_box = gui.widgetBox(script_tab, "Python script", addSpace=True, orientation="horizontal")
        script_box.layout().addWidget(self.shadow4_script)

        view_box = oasysgui.widgetBox(plot_tab, "", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.view_type_combo = gui.comboBox(view_box_1, self, "view_type", label="Plot Results",
                                            labelWidth=220,
                                            items=["No", "Yes"],
                                            callback=self.set_plot_quality, sendSelectedValue=False, orientation="horizontal")

        self.tabs = oasysgui.tabWidget(plot_tab)

        self.tabs_setting = oasysgui.tabWidget(self.controlArea)
        self.tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_bas = oasysgui.createTabPage(self.tabs_setting, "General")

        box_1 = oasysgui.widgetBox(tab_bas, "Calculation Parameters", addSpace=True, orientation="vertical", height=130)

        gui.comboBox(box_1, self, "calculation_type", label="Calculation", labelWidth=70,
                     items=["Diffraction by Simple Aperture",
                            "Diffraction by Mirror or Grating Size",
                            "Diffraction by Mirror Size + Figure Errors",
                            "Diffraction by Grating Size + Figure Errors",
                            "Diffraction by Lens Size",
                            "Diffraction by Lens Size + Thickness Errors",
                            "Diffraction by KB Size",
                            "Diffraction by KB Size + Figure Errors"],
                     callback=self.set_calculation_type,
                     sendSelectedValue=False, orientation="horizontal")

        self.cb_diffraction_plane = gui.comboBox(box_1, self, "diffraction_plane", label="Diffraction Plane", labelWidth=310,
                                                 items=["Sagittal", "Tangential", "Both (2D)", "Both (1D+1D)"],
                                                 callback=self.set_diffraction_plane,
                                                 sendSelectedValue=False, orientation="horizontal")

        self.cb_propagation_type = gui.comboBox(box_1, self, "propagation_type", label="Propagation Type", labelWidth=310,
                                                items=["Far Field", "Near Field", "Both"],
                                                sendSelectedValue=False, orientation="horizontal", callback=self.set_propagation_type)

        gui.separator(box_1)

        box_2 = oasysgui.widgetBox(tab_bas, "Numerical Control Parameters", addSpace=True, orientation="vertical", height=140)

        self.le_n_bins_x = oasysgui.lineEdit(box_2, self, "n_bins_x", "Number of bins for I(Sagittal) histogram", labelWidth=260, valueType=int, orientation="horizontal")
        self.le_n_bins_z = oasysgui.lineEdit(box_2, self, "n_bins_z", "Number of bins for I(Tangential) histogram", labelWidth=260, valueType=int, orientation="horizontal")
        self.le_n_peaks   = oasysgui.lineEdit(box_2, self, "n_peaks", "Number of diffraction peaks", labelWidth=260, valueType=int, orientation="horizontal")
        self.le_fft_n_pts = oasysgui.lineEdit(box_2, self, "fft_n_pts", "Number of points for FFT", labelWidth=260, valueType=int, orientation="horizontal")

        box_4 = oasysgui.widgetBox(tab_bas, "Calculation Congruence Parameters", addSpace=True, orientation="vertical", height=100)

        self.cb_analyze_geometry = gui.comboBox(box_4, self, "analyze_geometry", label="Analize geometry to avoid unuseful calculations", labelWidth=310,
                                                items=["No", "Yes"],
                                                sendSelectedValue=False, orientation="horizontal")


        gui.comboBox(box_4, self, "send_original_beam", label="Send Original Beam in case of failure", labelWidth=310,
                     items=["No", "Yes"],
                     sendSelectedValue=False, orientation="horizontal")

        self.set_calculation_type()

        self.initializeTabs()

        self.shadow_output = oasysgui.textArea(height=580, width=800)

        out_box = gui.widgetBox(out_tab, "System Output", addSpace=True, orientation="horizontal")
        out_box.layout().addWidget(self.shadow_output)

        self.set_plot_quality()

    def set_plot_quality(self):
        self.progressBarInit()

        if not self.__plotted_data is None:
            try:
                self.initializeTabs()
                if self.view_type == 1: self.plot_results(self.__plotted_data)

            except Exception as exception:
                QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

                if self.IS_DEVELOP: raise exception
        else:
            self.initializeTabs()

        self.progressBarFinished()

    def initializeTabs(self):
        self.tabs.clear()

        if self.diffraction_plane in [HybridDiffractionPlane.SAGITTAL, HybridDiffractionPlane.TANGENTIAL]:
            if self.propagation_type == HybridPropagationType.BOTH:
                self.tab = [gui.createTabPage(self.tabs, u"\u2206" + "Divergence (Far Field)"),
                            gui.createTabPage(self.tabs, u"\u2206" + "Position (Near Field)"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Far Field)"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Near Field)")]
            elif self.propagation_type == HybridPropagationType.FAR_FIELD:
                self.tab = [gui.createTabPage(self.tabs, u"\u2206" + "Divergence (Far Field)"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Far Field)")]
            elif self.propagation_type == HybridPropagationType.NEAR_FIELD:
                self.tab = [gui.createTabPage(self.tabs, u"\u2206" + "Position (Near Field)"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Near Field)")]
        elif self.diffraction_plane == HybridDiffractionPlane.BOTH_2D:
             self.tab = [gui.createTabPage(self.tabs, u"\u2206" + "Divergence (Far Field)"),
                         gui.createTabPage(self.tabs, "Spatial Distribution (Far Field)")]
        elif self.diffraction_plane == HybridDiffractionPlane.BOTH_2X1D:
            if self.propagation_type == HybridPropagationType.BOTH:
                self.tab = [gui.createTabPage(self.tabs, u"\u2206" + "Divergence (Far Field (S))"),
                            gui.createTabPage(self.tabs, u"\u2206" + "Position (Near Field (S))"),
                            gui.createTabPage(self.tabs, u"\u2206" + "Divergence (Far Field (T))"),
                            gui.createTabPage(self.tabs, u"\u2206" + "Position (Near Field (T))"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Far Field)"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Near Field)")]
            elif self.propagation_type == HybridPropagationType.FAR_FIELD:
                self.tab = [gui.createTabPage(self.tabs, u"\u2206" + "Divergence (Far Field (S))"),
                            gui.createTabPage(self.tabs, u"\u2206" + "Divergence (Far Field (T))"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Far Field)")]
            elif self.propagation_type == HybridPropagationType.NEAR_FIELD:
                self.tab = [gui.createTabPage(self.tabs, u"\u2206" + "Position (Near Field (S))"),
                            gui.createTabPage(self.tabs, u"\u2206" + "Position (Near Field (T))"),
                            gui.createTabPage(self.tabs, "Spatial Distribution (Near Field)")]

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

        self.plot_canvas = [None, None, None, None, None, None]

    def plot_xy(self, beam_out, progressBarValue, var_x, var_y, plot_canvas_index, title, xtitle, ytitle, xum="", yum=""):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = ShadowPlot.DetailedPlotWidget()
            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        self.plot_canvas[plot_canvas_index].plot_xy(beam_out, var_x, var_y, title, xtitle, ytitle, xum=xum, yum=yum)

        self.progressBarSet(progressBarValue)

    def plot_histo(self, beam_out, progressBarValue, var, plot_canvas_index, title, xtitle, ytitle, xum=""):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = ShadowPlot.DetailedHistoWidget()
            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        self.plot_canvas[plot_canvas_index].plot_histo(beam_out, var, 1, None, 23, title, xtitle, ytitle, xum=xum)

        self.progressBarSet(progressBarValue)

    def plot_emtpy(self, progressBarValue, plot_canvas_index):
        if self.plot_canvas[plot_canvas_index] is None:
            widget = QWidget()
            widget.setLayout(QHBoxLayout())
            label = QLabel(widget)
            label.setPixmap(QPixmap(QImage(os.path.join(resources.package_dirname("orangecontrib.shadow.widgets.special_elements"), "../../../../../OASYS1-SHADOW4/orangecontrib/shadow4/widgets/optics/icons", "no_result.png"))))

            widget.layout().addWidget(label)

            self.plot_canvas[plot_canvas_index] = widget

            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        self.progressBarSet(progressBarValue)

    def plot_xy_hybrid(self, progressBarValue, scaled_matrix, plot_canvas_index, title, xtitle, ytitle, var1, var2):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = ImageView()
            self.plot_canvas[plot_canvas_index].setColormap({"name":"temperature", "normalization":"linear", "autoscale":True, "vmin":0, "vmax":0, "colors":256})
            self.plot_canvas[plot_canvas_index].setMinimumWidth(590)
            self.plot_canvas[plot_canvas_index].setMaximumWidth(590)

            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        factor1 = ShadowPlot.get_factor(var1) # results are in m
        factor2 = ShadowPlot.get_factor(var2) # results are in m

        xmin, xmax = min(scaled_matrix.x_coord), max(scaled_matrix.x_coord)
        ymin, ymax = min(scaled_matrix.y_coord), max(scaled_matrix.y_coord)

        dim_x, dim_y = scaled_matrix.shape()

        origin = (xmin*factor1, ymin*factor2)
        scale = (abs((xmax-xmin)/dim_x)*factor1, abs((ymax-ymin)/dim_y)*factor2)

        data_to_plot = []
        for y_index in range(0, dim_y):
            x_values = []
            for x_index in range(0, dim_x):
                x_values.append(scaled_matrix.z_values[x_index, y_index])

            data_to_plot.append(x_values)

        self.plot_canvas[plot_canvas_index].setImage(numpy.array(data_to_plot), origin=origin, scale=scale)
        self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
        self.plot_canvas[plot_canvas_index].setGraphTitle(title)

        self.progressBarSet(progressBarValue)

    def plot_histo_hybrid(self, progressBarValue, scaled_array, plot_canvas_index, title, xtitle, ytitle, var):
        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow(roi=False, control=False, position=True)
            self.plot_canvas[plot_canvas_index].setDefaultPlotLines(True)
            self.plot_canvas[plot_canvas_index].setActiveCurveStyle(color='blue')
            self.plot_canvas[plot_canvas_index].setInteractiveMode(mode='zoom')

            self.tab[plot_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

        factor = ShadowPlot.get_factor(var) # results are in m

        self.plot_canvas[plot_canvas_index].addCurve(scaled_array.scale*factor, scaled_array.np_array, "crv_"+ytitle, symbol='', color="blue", replace=True) #'+', '^', ','
        self.plot_canvas[plot_canvas_index]._backend.ax.get_yaxis().get_major_formatter().set_useOffset(True)
        self.plot_canvas[plot_canvas_index]._backend.ax.get_yaxis().get_major_formatter().set_scientific(True)
        self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
        self.plot_canvas[plot_canvas_index].setGraphTitle(title)
        self.plot_canvas[plot_canvas_index].replot()

        self.progressBarSet(progressBarValue)

    @Inputs.shadow_data
    def set_shadow_data(self, input_data: ShadowData):
        if ShadowCongruence.check_empty_data(input_data):
            if ShadowCongruence.check_empty_beam(input_data.beam):
                self.__input_data = input_data.duplicate()

                if self.is_automatic_run: self.run_hybrid()

    def set_calculation_type(self):
        self.cb_diffraction_plane.setEnabled(True)
        self.cb_propagation_type.setEnabled(True)
        self.cb_analyze_geometry.setEnabled(True)

        if self.tabs_setting.count() == 2: self.tabs_setting.removeTab(1)

        if self.calculation_type == HybridCalculationType.SIMPLE_APERTURE:
            self.cb_propagation_type.setEnabled(False)
            self.propagation_type = HybridPropagationType.FAR_FIELD
        elif self.calculation_type in [HybridCalculationType.CRL_SIZE, HybridCalculationType.CRL_SIZE_AND_ERROR_PROFILE]:
            self.diffraction_plane = HybridDiffractionPlane.BOTH_2D
            self.propagation_type = HybridPropagationType.FAR_FIELD
            self.cb_diffraction_plane.setEnabled(False)
            self.cb_propagation_type.setEnabled(False)
            if self.calculation_type == HybridCalculationType.CRL_SIZE_AND_ERROR_PROFILE: self.create_tab_thickness()
        elif self.calculation_type in [HybridCalculationType.KB_SIZE, HybridCalculationType.KB_SIZE_AND_ERROR_PROFILE]:
            self.diffraction_plane = HybridDiffractionPlane.BOTH_2X1D
            self.cb_diffraction_plane.setEnabled(False)

        if self.diffraction_plane == HybridDiffractionPlane.BOTH_2D:
            self.cb_propagation_type.setEnabled(False)
            self.propagation_type = HybridPropagationType.FAR_FIELD

        if self.calculation_type in [HybridCalculationType.MIRROR_SIZE_AND_ERROR_PROFILE,
                                     HybridCalculationType.GRATING_SIZE_AND_ERROR_PROFILE,
                                     HybridCalculationType.CRL_SIZE_AND_ERROR_PROFILE,
                                     HybridCalculationType.KB_SIZE_AND_ERROR_PROFILE]:
            self.analyze_geometry = 0
            self.cb_analyze_geometry.setEnabled(False)

        self.set_diffraction_plane()
        self.set_propagation_type()

    def set_diffraction_plane(self):
        self.le_n_bins_x.setEnabled(self.diffraction_plane in [HybridDiffractionPlane.SAGITTAL, HybridDiffractionPlane.BOTH_2X1D, HybridDiffractionPlane.BOTH_2D])
        self.le_n_bins_z.setEnabled(self.diffraction_plane in [HybridDiffractionPlane.TANGENTIAL, HybridDiffractionPlane.BOTH_2X1D, HybridDiffractionPlane.BOTH_2D])

    def set_propagation_type(self): pass # maybe it can be useful again


    # --------------------------------------------------
    # HybridListener methods
    # --------------------------------------------------

    def status_message(self, message : str): self.setStatusMessage(message)
    def set_progress_value(self, value):
        if value >= 100: self.progressBarFinished()
        elif value <= 0: self.progressBarInit()
        else:            self.progressBarSet(value)
    def warning_message(self, message : str): QMessageBox.warning(self, "Warning", str(message), QMessageBox.Ok)
    def error_message(self, message : str):   QMessageBox.critical(self, "Error", str(message), QMessageBox.Ok)

    def run_hybrid(self):
        try:
            self.setStatusMessage("")
            self.progressBarInit()
            self.initializeTabs()

            if ShadowCongruence.check_empty_data(self.__input_data):
                if ShadowCongruence.check_empty_beam(self.__input_data.beam):
                    sys.stdout = EmittingStream(textWritten=self.write_stdout)

                    self.check_fields()

                    beam     = self.__input_data.beam
                    beamline = self.__input_data.beamline

                    additional_parameters = {}

                    if self.calculation_type == HybridCalculationType.CRL_SIZE_AND_ERROR_PROFILE:
                        additional_parameters["crl_error_profiles"] = self.crl_error_profiles
                        if self.crl_material_data == 0: additional_parameters["crl_material"] = self.crl_material
                        else:                           additional_parameters["crl_delta"]    = self.crl_delta
                        additional_parameters["crl_scaling_factor"] = self.crl_scaling_factor
                        additional_parameters["crl_coords_to_m"]    = self.crl_coords_to_m
                        additional_parameters["crl_thickness_to_m"] = self.crl_thickness_to_m

                    if self.calculation_type in [HybridCalculationType.KB_SIZE, HybridCalculationType.KB_SIZE_AND_ERROR_PROFILE]:
                        if beamline.get_beamline_elements_number() < 2: raise ValueError("Calculation with KB not possible: not enough elements in the beamline")

                        kb_mirror_1_element = beamline.get_beamline_element_at(-2)
                        kb_mirror_2_element = beamline.get_beamline_element_at(-1)
                        beam             = S4HybridBeam(beam=[kb_mirror_2_element.get_input_beam(), beam])
                        optical_element  = S4HybridOE(optical_element=[kb_mirror_1_element, kb_mirror_2_element])
                    else:
                        beamline_element = beamline.get_beamline_element_at(-1)
                        beam             = S4HybridBeam(beam=beam)
                        optical_element  = S4HybridOE(optical_element=beamline_element)

                    input_parameters = HybridInputParameters(listener=self,
                                                             beam=beam,
                                                             optical_element=optical_element,
                                                             diffraction_plane=self.diffraction_plane,
                                                             propagation_type=self.propagation_type,
                                                             n_bins_x=int(self.n_bins_x),
                                                             n_bins_z=int(self.n_bins_z),
                                                             n_peaks=int(self.n_peaks),
                                                             fft_n_pts=int(self.fft_n_pts),
                                                             analyze_geometry=self.analyze_geometry==1,
                                                             random_seed=None,  # TODO: add field
                                                             **additional_parameters)
                    try:

                        beamline = self.__input_data.beamline.duplicate()
                        element = S4HybridScreenElement(hybrid_screen=S4HybridScreen(calculation_type=self.calculation_type),
                                                        hybrid_input_parameters=input_parameters)

                        beamline.append_beamline_element(element)

                        #
                        # script
                        #
                        script = beamline.to_python_code()
                        script += "\n\n\n# test plot"
                        script += "\nif True:"
                        script += "\n   from srxraylib.plot.gol import plot_scatter"
                        script += "\n   plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1), title='(Intensity,Photon Energy)', plot_histograms=0)"
                        script += "\n   plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1), title='(X,Z) in microns')"
                        self.shadow4_script.set_code(script)

                        self.progressBarInit()
                        beam, mirr, calculation_result = element.trace_beam()

                        # PARAMETERS IN SET MODE
                        self.n_bins_x                                   = int(input_parameters.n_bins_x)
                        self.n_bins_z                                   = int(input_parameters.n_bins_z)
                        self.n_peaks                                    = int(input_parameters.n_peaks)
                        self.fft_n_pts                                  = int(input_parameters.fft_n_pts)

                        self.__plotted_data = calculation_result
                        if self.view_type==1: self.plot_results(beam, calculation_result)

                        if self.propagation_type == HybridPropagationType.BOTH:
                            shadow_data_ff = ShadowData(beam=beam[0], beamline=beamline)
                            shadow_data_ff.scanning_data = self.__input_data.scanning_data
                            shadow_data_ff.initial_flux  = self.__input_data.initial_flux
                            shadow_data_nf = ShadowData(beam=beam[1], beamline=beamline)
                            shadow_data_nf.scanning_data = self.__input_data.scanning_data
                            shadow_data_nf.initial_flux  = self.__input_data.initial_flux
                            self.Outputs.shadow_data_ff.send(shadow_data_ff)
                            self.Outputs.shadow_data_nf.send(shadow_data_nf)
                        elif self.propagation_type == HybridPropagationType.FAR_FIELD:
                            shadow_data_ff = ShadowData(beam=beam, beamline=beamline)
                            shadow_data_ff.scanning_data = self.__input_data.scanning_data
                            shadow_data_ff.initial_flux  = self.__input_data.initial_flux
                            self.Outputs.shadow_data_ff.send(shadow_data_ff)
                        elif self.propagation_type == HybridPropagationType.NEAR_FIELD:
                            shadow_data_nf = ShadowData(beam=beam, beamline=beamline)
                            shadow_data_nf.scanning_data = self.__input_data.scanning_data
                            shadow_data_nf.initial_flux  = self.__input_data.initial_flux
                            self.Outputs.shadow_data_nf.send(shadow_data_nf)

                        self.Outputs.trigger.send(TriggerIn(new_object=True))
                    except Exception as e:
                        if self.send_original_beam==1:
                            self.Outputs.shadow_data_ff.send(self.__input_data.duplicate())
                            self.Outputs.trigger.send(TriggerIn(new_object=True))
                        else:
                            raise e
                else:
                    raise Exception("Empty Input Beam")
            else:
                raise Exception("Empty Input Data")
        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

        self.setStatusMessage("")
        self.progressBarFinished()

    def plot_results(self, beam, calculation_result : HybridCalculationResult):
        geometry_analysis = calculation_result.geometry_analysis

        do_plot_sagittal   = True
        do_plot_tangential = True

        if not self.calculation_type in [HybridCalculationType.MIRROR_SIZE_AND_ERROR_PROFILE,
                                         HybridCalculationType.GRATING_SIZE_AND_ERROR_PROFILE,
                                         HybridCalculationType.CRL_SIZE_AND_ERROR_PROFILE] \
                and self.analyze_geometry == 1:
            do_plot_sagittal   = not geometry_analysis.has_result(HybridGeometryAnalysis.BEAM_NOT_CUT_SAGITTALLY)
            do_plot_tangential = not geometry_analysis.has_result(HybridGeometryAnalysis.BEAM_NOT_CUT_TANGENTIALLY)

        if do_plot_sagittal or do_plot_tangential: self.setStatusMessage("Plotting Results")
        
        def plot_direction(direction="S", do_plot=True, progress=[84, 88, 92, 96], start_index=0, plot_shadow=True):
            if direction=="S":
                ax         = "X"
                divergence = calculation_result.divergence_sagittal
                position   = calculation_result.position_sagittal
                var        = 1
            else:
                ax         = "Z"
                divergence = calculation_result.divergence_tangential
                position   = calculation_result.position_tangential
                var        = 3

            if do_plot:
                if self.propagation_type == HybridPropagationType.BOTH:
                    self.plot_histo_hybrid(progress[0], divergence, start_index + 0,     title=u"\u2206" + ax + "p", xtitle=r'$\Delta$' + ax + 'p [$\\mu$rad]', ytitle='Arbitrary Units', var=4)
                    self.plot_histo_hybrid(progress[2], position,   start_index + 1, title=u"\u2206" + ax, xtitle=r'$\Delta$' + ax + ' [$\\mu$m]', ytitle='Arbitrary Units', var=1)
                    if plot_shadow:
                        self.plot_histo(beam[0], progress[1], var, plot_canvas_index=start_index + 2, title=ax, xtitle=r'' + ax + ' [$\\mu$m]', ytitle=r'Number of Rays', xum=("X [" + u"\u03BC" + "m]"))
                        self.plot_histo(beam[1], progress[3], var, plot_canvas_index=start_index + 3, title=ax, xtitle=r'' + ax + ' [$\\mu$m]', ytitle=r'Number of Rays', xum=("X [" + u"\u03BC" + "m]"))
                elif self.propagation_type == HybridPropagationType.FAR_FIELD:
                    self.plot_histo_hybrid(progress[1], divergence, start_index + 0, title=u"\u2206" + ax + "p", xtitle=r'$\Delta$' + ax + 'p [$\\mu$rad]', ytitle=r'Arbitrary Units', var=4)
                    if plot_shadow:
                        self.plot_histo(beam, progress[3], var, plot_canvas_index=start_index + 1, title=ax, xtitle=r'' + ax + ' [$\\mu$m]', ytitle=r'Number of Rays', xum=("X [" + u"\u03BC" + "m]"))
                elif self.propagation_type == HybridPropagationType.NEAR_FIELD:
                    self.plot_histo_hybrid(progress[1], position, start_index + 0, title=u"\u2206" + ax, xtitle=r'$\Delta$' + ax + ' [$\\mu$m]', ytitle=r'Arbitrary Units', var=1)
                    if plot_shadow:
                        self.plot_histo(beam, progress[3], var, plot_canvas_index=start_index + 1, title=ax, xtitle=r'' + ax + ' [$\\mu$m]', ytitle=r'Number of Rays', xum=("X [" + u"\u03BC" + "m]"))
            else:
                if self.propagation_type == HybridPropagationType.BOTH:
                    self.plot_emtpy(progress[0], start_index + 0)
                    self.plot_emtpy(progress[1], start_index + 1)
                    if plot_shadow:
                        self.plot_emtpy(progress[2], start_index + 2)
                        self.plot_emtpy(progress[3], start_index + 3)
                else:
                    self.plot_emtpy(progress[1], start_index + 0)
                    if plot_shadow:
                        self.plot_emtpy(progress[3], start_index + 1)

        if   self.diffraction_plane == HybridDiffractionPlane.SAGITTAL:   plot_direction("S", do_plot_sagittal)
        elif self.diffraction_plane == HybridDiffractionPlane.TANGENTIAL: plot_direction("T", do_plot_tangential)
        elif self.diffraction_plane == HybridDiffractionPlane.BOTH_2D:
            if do_plot_sagittal and do_plot_tangential:
                self.plot_xy_hybrid(88, calculation_result.divergence_2D, plot_canvas_index=0, title="X',Z'",xtitle="X' [$\\mu$rad]", ytitle="Z' [$\\mu$rad]", var1=4, var2=6)
                self.plot_xy(beam, 96, 1, 3, plot_canvas_index=1, title="X,Z", xtitle=r'X [$\mu$m]', ytitle=r'Z [$\mu$m]', xum=("X [" + u"\u03BC" + "m]"), yum=("Z [" + u"\u03BC" + "m]"))
            else:
                if do_plot_sagittal:
                    self.plot_histo_hybrid(88, calculation_result.divergence_sagittal, plot_canvas_index=0, title=u"\u2206" + "X'", xtitle="$\\Delta$X' [$\\mu$rad]", ytitle=r'Arbitrary Units', var=4)
                    self.plot_histo(beam, 96, 1, plot_canvas_index=1, title="X", xtitle=r'X [$\mu$m]', ytitle=r'Number of Rays', xum=("X [" + u"\u03BC" + "m]"))
                elif do_plot_tangential:
                    self.plot_histo_hybrid(88, calculation_result.divergence_tangential, 0, title=u"\u2206" + "Z'", xtitle="$\\Delta$Z' [$\\mu$rad]", ytitle=r'Arbitrary Units', var=6)
                    self.plot_histo(beam, 96, 3, plot_canvas_index=1, title="Z", xtitle=r'Z [$\mu$m]', ytitle=r'Number of Rays', xum=("Z [" + u"\u03BC" + "m]"))
                else:
                    self.plot_emtpy(88, 0)
                    self.plot_emtpy(96, 1)
        elif self.diffraction_plane == HybridDiffractionPlane.BOTH_2X1D:
            plot_direction("S", do_plot_sagittal,   progress=[82, 82, 84, 84], plot_shadow=False)
            plot_direction("T", do_plot_tangential, progress=[86, 88, 94, 98], start_index=2 if self.propagation_type == HybridPropagationType.BOTH else 1, plot_shadow=False)

            if self.propagation_type == HybridPropagationType.BOTH:
                self.plot_xy(beam[0], 94, 1, 3, plot_canvas_index=4 if self.propagation_type == HybridPropagationType.BOTH else 2,
                             title="X,Z", xtitle=r'X [$\mu$m]', ytitle=r'Z [$\mu$m]', xum=("X [" + u"\u03BC" + "m]"), yum=("Z [" + u"\u03BC" + "m]"))
                self.plot_xy(beam[1], 94, 1, 3, plot_canvas_index=5 if self.propagation_type == HybridPropagationType.BOTH else 3,
                             title="X,Z", xtitle=r'X [$\mu$m]', ytitle=r'Z [$\mu$m]', xum=("X [" + u"\u03BC" + "m]"), yum=("Z [" + u"\u03BC" + "m]"))
            else:
                self.plot_xy(beam, 94, 1, 3, plot_canvas_index=2,
                             title="X,Z", xtitle=r'X [$\mu$m]', ytitle=r'Z [$\mu$m]', xum=("X [" + u"\u03BC" + "m]"), yum=("Z [" + u"\u03BC" + "m]"))

    def check_fields(self):
        if self.diffraction_plane in [HybridDiffractionPlane.SAGITTAL, HybridDiffractionPlane.BOTH_2D]:
            congruence.checkStrictlyPositiveNumber(self.n_bins_x, "Number of bins for I(Sagittal) histogram")
        if self.diffraction_plane in [HybridDiffractionPlane.TANGENTIAL, HybridDiffractionPlane.BOTH_2D]:
            congruence.checkStrictlyPositiveNumber(self.n_bins_z, "Number of bins for I(Tangential) histogram")

        congruence.checkStrictlyPositiveNumber(self.n_peaks, "Number of diffraction peaks")
        congruence.checkStrictlyPositiveNumber(self.fft_n_pts, "Number of points for FFT")

        if self.calculation_type == HybridCalculationType.CRL_SIZE_AND_ERROR_PROFILE:
            if len(self.crl_error_profiles) == 0: raise ValueError("No Thickness error profile specified")
            if self.crl_material_data == 0: self.crl_material = congruence.checkEmptyString(self.crl_material, "Chemical Formula")
            else: congruence.checkStrictlyPositiveNumber(self.crl_delta, "Refractive Index (\u03b4)")
            congruence.checkPositiveNumber(self.crl_scaling_factor, "Thickness Error Scaling Factor")


    def write_stdout(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

    @Inputs.trigger
    def set_trigger_parameters_for_optics(self, trigger):
        super(HybridScreen, self).set_trigger_parameters_for_optics(trigger)

    @Inputs.thickness_error_data
    def set_thickness_error_profiles(self, thickness_errors_data):
        try:
            thickness_error_profile_data_files = thickness_errors_data.thickness_error_profile_data_files

            if not thickness_error_profile_data_files is None:
                self.crl_error_profiles = [thickness_error_file for thickness_error_file in thickness_error_profile_data_files]
                if self.calculation_type==HybridCalculationType.CRL_SIZE_AND_ERROR_PROFILE: self.refresh_files_text_area()
        except Exception as exception:
            QMessageBox.critical(self, "Error", exception.args[0], QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

    def refresh_files_text_area(self):
        text = ""
        for file in self.crl_error_profiles: text += file + "\n"

        self.files_area.setText(text)

    def create_tab_thickness(self):
        tab_thick = oasysgui.createTabPage(self.tabs_setting, "Thickness Error")

        input_box = oasysgui.widgetBox(tab_thick, "Thickness Error Files", addSpace=True, orientation="vertical", height=450, width=self.CONTROL_AREA_WIDTH-20)

        gui.comboBox(input_box, self, "crl_material_data", label="Material Properties from", labelWidth=180,
                     items=["Chemical Formula", "Absorption Parameters"],
                     callback=self.set_crl_material_data,
                     sendSelectedValue=False, orientation="horizontal")

        self.input_box_1 = oasysgui.widgetBox(input_box, "", addSpace=False, orientation="vertical", width=self.CONTROL_AREA_WIDTH-40)
        self.input_box_2 = oasysgui.widgetBox(input_box, "", addSpace=False, orientation="vertical", width=self.CONTROL_AREA_WIDTH-40)

        oasysgui.lineEdit(self.input_box_1, self, "crl_material", "Chemical Formula", labelWidth=260, valueType=str, orientation="horizontal")
        oasysgui.lineEdit(self.input_box_2, self, "crl_delta", "Refractive Index (\u03b4)", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_crl_material_data()

        self.files_area = oasysgui.textArea(height=265)

        input_box.layout().addWidget(self.files_area)

        self.refresh_files_text_area()

        oasysgui.lineEdit(input_box, self, "crl_thickness_to_m", label="Thickness conversion to m", labelWidth=260, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(input_box, self, "crl_coords_to_m",   label="Coordinates conversion to m", labelWidth=260, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(input_box, self, "crl_scaling_factor", "Thickness Error Scaling Factor", labelWidth=260, valueType=float, orientation="horizontal")

    def set_crl_material_data(self):
        self.input_box_1.setVisible(self.crl_material_data==0)
        self.input_box_2.setVisible(self.crl_material_data==1)

add_widget_parameters_to_module(__name__)


import sys, numpy

from AnyQt.QtWidgets import QMessageBox

from matplotlib import rcParams

from silx.gui.plot import Plot2D

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from oasys2.widget.widget import OWAction
from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.util.widget_util import EmittingStream
from oasys2.widget.gui import Styles
from oasys2.widget.util.widget_objects import TriggerIn
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from syned.beamline.element_coordinates import ElementCoordinates

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, TriggerToolsDecorator
from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from shadow4_advanced.fresnel_zone_plate.s4_fresnel_zone_plate import (S4FresnelZonePlate, S4FresnelZonePlateElement,
                                                                       FZPCalculationInputParameters, FZPAttributes, FZPSimulatorOptions, FZPCalculationResult)
GOOD = 1

COLLIMATED_SOURCE_LIMIT = 1e4 #m

class FresnelZonePlate(GenericElement, TriggerToolsDecorator):
    name = "Hybrid Fresnel Zone Plate"
    description = "Advanced: Fresnel Zone Plate"
    icon = "icons/zone_plate.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 4
    category = "Optical Elements"
    keywords = ["data", "file", "load", "read"]

    class Inputs:
        shadow_data = Input("Shadow Data", ShadowData, default=True, auto_summary=False)
        trigger     = TriggerToolsDecorator.get_trigger_input()

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)
        trigger     = TriggerToolsDecorator.get_trigger_output()

    input_data = None
    output_data = None

    NONE_SPECIFIED = "NONE SPECIFIED"

    ONE_ROW_HEIGHT = 65
    TWO_ROW_HEIGHT = 110
    THREE_ROW_HEIGHT = 170

    INNER_BOX_WIDTH_L3=322
    INNER_BOX_WIDTH_L2=335
    INNER_BOX_WIDTH_L1=358
    INNER_BOX_WIDTH_L0=375

    source_plane_distance = Setting(10.0)
    image_plane_distance  = Setting(20.0)

    height = Setting(400.0) # nm
    diameter = Setting(50.0) # um
    b_min = Setting(50.0) # nm
    zone_plate_material = Setting('Au')
    template_material = Setting('SiO2')

    zone_plate_type = Setting(0)
    width_coating = Setting(20) # nm
    height1_factor = Setting(0.33)
    height2_factor = Setting(0.67)

    with_central_stop = Setting(0)
    cs_diameter = Setting(10.0) # um

    with_order_sorting_aperture = Setting(0)

    osa_position = Setting(10.0) # user units
    osa_diameter =  Setting(30.0) # um

    source_distance_flag = Setting(0)
    source_distance = Setting(0.0)

    image_distance_flag = Setting(1)
    image_distance = Setting(0.0)

    multipool = Setting(1)

    with_multi_slicing = Setting(0)
    n_slices = Setting(100)

    increase_resolution = Setting(1)
    increase_points = Setting(200)

    n_points = Setting(5000)
    last_index = Setting(100)

    ##################################################

    avg_energy = 0.0
    number_of_zones = 0
    focal_distance = 0.0
    efficiency = 0.0

    def __init__(self):
        super(FresnelZonePlate, self).__init__(show_automatic_box=True, has_footprint=False)

        self.runaction = OWAction("Run Shadow4/Trace", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run Shadow4/Trace", callback=self.run_shadow4)
        button.setStyleSheet(Styles.button_blue)

        button = gui.button(button_box, self, "Reset Fields", callback=self.callResetSettings)
        button.setStyleSheet(Styles.button_red)

        gui.separator(self.controlArea)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_pos = oasysgui.createTabPage(tabs_setting, "Position")

        upper_box = oasysgui.widgetBox(tab_pos, "Optical Element Orientation", addSpace=True, orientation="vertical")

        self.le_source_plane_distance = oasysgui.lineEdit(upper_box, self, "source_plane_distance", "Source Plane Distance [m]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_image_plane_distance  = oasysgui.lineEdit(upper_box, self, "image_plane_distance", "Image Plane Distance [m]", labelWidth=260, valueType=float, orientation="horizontal")

        tab_bas = oasysgui.createTabPage(tabs_setting, "Fresnel Zone Plate")

        ##########################################
        ##########################################
        # BASIC SETTINGS
        ##########################################
        ##########################################

        tabs_basic_setting = oasysgui.tabWidget(tab_bas)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT-5)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_zone_plate_1 = oasysgui.createTabPage(tabs_basic_setting, "Input Parameters")
        tab_zone_plate_2 = oasysgui.createTabPage(tabs_basic_setting, "Propagation Parameters")
        tab_zone_plate_3 = oasysgui.createTabPage(tabs_basic_setting, "Output Parameters")

        zp_box = oasysgui.widgetBox(tab_zone_plate_1, "F.Z.P. Parameters", addSpace=False, orientation="vertical", height=475)

        oasysgui.lineEdit(zp_box, self, "b_min",  "Outermost Zone Width/Period [nm]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(zp_box, self, "diameter", "F.Z.P. Diameter [" + u"\u03BC" + "m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(zp_box, self, "height",  "F.Z.P. Height [nm]", labelWidth=260, valueType=float, orientation="horizontal")

        gui.comboBox(zp_box, self, "zone_plate_type", label="Type of F.Z.P.", labelWidth=350,
                     items=["Ordinary", "Zone-Doubled", "Zone-Filled", "Two-Level"],
                     callback=self.set_FZPType, sendSelectedValue=False, orientation="horizontal")

        oasysgui.lineEdit(zp_box, self, "zone_plate_material",  "F.Z.P. Material", labelWidth=160, valueType=str, orientation="horizontal")

        self.ord_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)

        self.zd_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        oasysgui.lineEdit(self.zd_box, self, "template_material",  "Template Material", labelWidth=160, valueType=str, orientation="horizontal")
        oasysgui.lineEdit(self.zd_box, self, "width_coating",  "Coating Width [nm]", labelWidth=260, valueType=float, orientation="horizontal")

        self.zf_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        oasysgui.lineEdit(self.zf_box, self, "template_material", "Template Material", labelWidth=160, valueType=str, orientation="horizontal")
        oasysgui.lineEdit(self.zf_box, self, "width_coating", "Coating Width [nm]", labelWidth=260, valueType=float, orientation="horizontal")

        self.tl_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        oasysgui.lineEdit(self.tl_box, self, "height1_factor",  "Height 1 Factor", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.tl_box, self, "height2_factor",  "Height 2 Factor", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_FZPType()

        gui.comboBox(zp_box, self, "with_central_stop", label="With Central Stop", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_WithCentralStop, sendSelectedValue=False, orientation="horizontal")

        self.cs_box_1 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)
        self.cs_box_2 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)

        oasysgui.lineEdit(self.cs_box_1, self, "cs_diameter", "C.S. Diameter [" + u"\u03BC" + "m]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_WithCentralStop()

        gui.comboBox(zp_box, self, "with_order_sorting_aperture", label="With Order Sorting Aperture", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_WithOrderSortingAperture, sendSelectedValue=False, orientation="horizontal")

        self.osa_box_1 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        self.osa_box_2 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)

        self.le_osa_position = oasysgui.lineEdit(self.osa_box_1, self, "osa_position", "O.S.A. position [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.osa_box_1, self, "osa_diameter", "O.S.A. Diameter [" + u"\u03BC" + "m]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_WithOrderSortingAperture()

        gui.comboBox(zp_box, self, "source_distance_flag", label="Source Distance", labelWidth=350,
                     items=["Same as Source Plane", "Different"],
                     callback=self.set_SourceDistanceFlag, sendSelectedValue=False, orientation="horizontal")

        self.zp_box_1 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)
        self.zp_box_2 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)

        self.le_source_distance = oasysgui.lineEdit(self.zp_box_1, self, "source_distance", "Source Distance [m]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_SourceDistanceFlag()

        gui.comboBox(zp_box, self, "image_distance_flag", label="Image Distance [m]", labelWidth=350,
                     items=["Image Plane Distance", "F.Z.P. Image Distance"],
                     callback=self.set_ImageDistanceFlag, sendSelectedValue=False, orientation="horizontal")

        self.set_ImageDistanceFlag()

        prop_box = oasysgui.widgetBox(tab_zone_plate_2, "Propagation Parameters", addSpace=False, orientation="vertical", height=270)

        oasysgui.lineEdit(prop_box, self, "n_points", "Nr. Sampling Points", labelWidth=260, valueType=int, orientation="horizontal")

        oasysgui.lineEdit(prop_box, self, "last_index", "Last Index of Focal Image", labelWidth=260, valueType=int, orientation="horizontal")

        gui.separator(prop_box)

        gui.comboBox(prop_box, self, "increase_resolution", label="Increase Resolution in Focal Image", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_IncreaseResolution, sendSelectedValue=False, orientation="horizontal")

        self.res_box_1 = oasysgui.widgetBox(prop_box, "", addSpace=False, orientation="vertical", height=30)
        self.res_box_2 = oasysgui.widgetBox(prop_box, "", addSpace=False, orientation="vertical", height=30)

        oasysgui.lineEdit(self.res_box_1, self, "increase_points", "Nr. Points", labelWidth=260, valueType=int, orientation="horizontal")

        self.set_IncreaseResolution()

        gui.separator(prop_box)

        gui.comboBox(prop_box, self, "multipool", label="Parallel Computing", labelWidth=350,
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal")


        zp_out_box = oasysgui.widgetBox(tab_zone_plate_3, "Output Parameters", addSpace=False, orientation="vertical", height=270)

        read_only_style = Styles.line_edit_read_only

        self.le_avg_wavelength = oasysgui.lineEdit(zp_out_box, self, "avg_energy", "Average Energy [eV]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_avg_wavelength.setReadOnly(True)
        self.le_avg_wavelength.setStyleSheet(read_only_style)

        self.le_number_of_zones = oasysgui.lineEdit(zp_out_box, self, "number_of_zones", "Number of Zones", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_number_of_zones.setReadOnly(True)
        self.le_number_of_zones.setStyleSheet(read_only_style)

        self.le_focal_distance = oasysgui.lineEdit(zp_out_box, self, "focal_distance", "Focal Distance [m]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_focal_distance.setReadOnly(True)
        self.le_focal_distance.setStyleSheet(read_only_style)

        self.le_image_distance = oasysgui.lineEdit(zp_out_box, self, "image_distance", "Image Distance (Q) [m]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_image_distance.setReadOnly(True)
        self.le_image_distance.setStyleSheet(read_only_style)

        self.le_efficiency = oasysgui.lineEdit(zp_out_box, self, "efficiency", "Efficiency % (Avg. Wavelength)", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_efficiency.setReadOnly(True)
        self.le_efficiency.setStyleSheet(read_only_style)


        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

        propagation_plot_tab = oasysgui.widgetBox(self.main_tabs, addToLayout=0, margin=4)

        self.main_tabs.insertTab(1, propagation_plot_tab, "TEMP")
        self.main_tabs.setTabText(0, "Shadow Plot")
        self.main_tabs.setTabText(1, "F.Z.P. Simulator Plot")

        self.prop_tabs = oasysgui.tabWidget(propagation_plot_tab)
        self.prop_tab = [oasysgui.createTabPage(self.prop_tabs, "Radial Intensity"),
                         oasysgui.createTabPage(self.prop_tabs, "Generated 2D distribution")]
        self.prop_plot_canvas = [None, None]

    def run_shadow4(self, scanning_data: ShadowData.ScanningData = None):
        if self.input_data is None:
            self.prompt_exception(ValueError("No input beam"))
            return
        if not scanning_data: scanning_data = None

        try:
            self.setStatusMessage("")
            self.progressBarInit()

            if ShadowCongruence.check_empty_data(self.input_data):
                if ShadowCongruence.check_good_beam(self.input_data.beam):
                    self.checkFields()

                    sys.stdout = EmittingStream(textWritten=self._write_stdout)

                    beamline = self.input_data.beamline.duplicate()

                    self.progressBarSet(10)

                    if self.source_distance_flag == 0: self.source_distance = self.source_plane_distance

                    input_parameters = FZPCalculationInputParameters(source_distance=self.source_distance,
                                                                     image_distance=self.image_plane_distance if self.image_distance_flag==0 else None,
                                                                     n_points=self.n_points,
                                                                     multipool=self.multipool==1,
                                                                     profile_last_index=self.last_index,
                                                                     increase_resolution=self.increase_resolution==1,
                                                                     increase_points=self.increase_points)
                    options = FZPSimulatorOptions(with_central_stop=self.with_central_stop==1,
                                                  cs_diameter=round(self.cs_diameter*1e-6, 7),
                                                  with_order_sorting_aperture=self.with_order_sorting_aperture==1,
                                                  osa_position=self.osa_position,
                                                  osa_diameter=round(self.osa_diameter*1e-6, 7),
                                                  zone_plate_type=self.zone_plate_type,
                                                  width_coating=round(self.width_coating*1e-9, 10),
                                                  height1_factor=self.height1_factor,
                                                  height2_factor=self.height2_factor,
                                                  with_range=False,
                                                  with_multi_slicing=self.with_multi_slicing==1,
                                                  n_slices=self.n_slices,
                                                  with_complex_amplitude=False,
                                                  store_partial_results=False)
                    attributes = FZPAttributes(height=round(self.height*1e-9, 10),
                                               diameter=round(self.diameter*1e-6, 7),
                                               b_min=round(self.b_min*1e-9, 10),
                                               zone_plate_material=self.zone_plate_material,
                                               template_material=self.template_material)

                    self.progressBarSet(30)

                    element = S4FresnelZonePlateElement(optical_element=S4FresnelZonePlate(input_parameters=input_parameters,
                                                                                           options=options,
                                                                                           attributes=attributes),
                                                        coordinates=ElementCoordinates(p=self.source_distance, q=self.image_distance),
                                                        input_beam=self.input_data.beam)
                    #element.set_movements(self.get_movements_instance()) for the future

                    print(element.info())

                    output_beam, calculation_result = element.trace_beam()

                    zone_plate_out = element.get_optical_element()

                    self.avg_energy      = 1e3*zone_plate_out.get_energy_in_KeV()
                    self.image_distance  = round(zone_plate_out.zp_image_distance, 6)
                    self.number_of_zones = zone_plate_out.get_n_zones()
                    self.focal_distance  = round(zone_plate_out.zp_focal_distance, 6)
                    self.efficiency      = round(calculation_result.efficiency*100, 2)

                    self.plot_propagation_results(calculation_result)

                    self.progressBarSet(80)

                    self.setStatusMessage("Plotting Results")

                    self._plot_results(output_beam, None)

                    self.setStatusMessage("")

                    beamline.append_beamline_element(element)

                    script = beamline.to_python_code()
                    script += "\n\n\n# test plot"
                    script += "\nif True:"
                    script += "\n   from srxraylib.plot.gol import plot_scatter"
                    script += "\n   plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1), title='(Intensity,Photon Energy)', plot_histograms=0)"
                    script += "\n   plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1), title='(X,Z) in microns')"
                    self.shadow4_script.set_code(script)

                    #
                    # send beam and trigger
                    #
                    output_data = ShadowData(beam=output_beam, beamline=beamline)
                    output_data.scanning_data = scanning_data

                    self.Outputs.shadow_data.send(output_data)
                    self.Outputs.trigger.send(TriggerIn(new_object=True))
                else:
                    raise Exception("Input Beam with no good rays")
            else:
                raise Exception("Empty Input Beam")

        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

        self.progressBarFinished()

    @Inputs.trigger
    def set_trigger_parameters_for_optics(self, trigger):
        super(FresnelZonePlate, self).set_trigger_parameters_for_optics(trigger)

    @Inputs.shadow_data
    def set_shadow_data(self, shadow_data):
        if ShadowCongruence.check_empty_data(shadow_data):
            self.input_data = shadow_data

            if self.is_automatic_run:
                self.run_shadow4()

    def checkFields(self):
        self.source_plane_distance = congruence.checkNumber(self.source_plane_distance, "Source plane distance")
        self.image_plane_distance = congruence.checkNumber(self.image_plane_distance, "Image plane distance")

    def callResetSettings(self):
        super()._call_reset_settings()

    def set_SourceMovement(self):
        self.sou_mov_box_1.setVisible(self.source_movement == 1)

    def set_MirrorMovement(self):
        self.mir_mov_box_1.setVisible(self.mirror_movement == 1)

    def set_FZPType(self):
        self.ord_box.setVisible(self.zone_plate_type == 0)
        self.zd_box.setVisible(self.zone_plate_type == 1)
        self.zf_box.setVisible(self.zone_plate_type == 2)
        self.tl_box.setVisible(self.zone_plate_type == 3)

    def set_SourceDistanceFlag(self):
        self.zp_box_1.setVisible(self.source_distance_flag == 1)
        self.zp_box_2.setVisible(self.source_distance_flag == 0)

    def set_WithCentralStop(self):
        self.cs_box_1.setVisible(self.with_central_stop == 1)
        self.cs_box_2.setVisible(self.with_central_stop == 0)

    def set_WithOrderSortingAperture(self):
        self.osa_box_1.setVisible(self.with_order_sorting_aperture == 1)
        self.osa_box_2.setVisible(self.with_order_sorting_aperture == 0)

    def set_WithMultislicing(self):
        self.ms_box_1.setVisible(self.with_multi_slicing == 1)
        self.ms_box_2.setVisible(self.with_multi_slicing == 0)

    def set_IncreaseResolution(self):
        self.res_box_1.setVisible(self.increase_resolution == 1)
        self.res_box_2.setVisible(self.increase_resolution == 0)

    def set_ImageDistanceFlag(self):
        self.le_image_plane_distance.setEnabled(self.image_distance_flag==0)

    def plot_propagation_results(self, calculation_result : FZPCalculationResult):
        self.plot_1D(0, calculation_result.radius*1e6, calculation_result.intensity_profile)
        self.plot_2D(1, calculation_result.xp * 1e6, calculation_result.zp * 1e6, calculation_result.dif_xpzp)

    def plot_1D(self, index, radius, profile_1D, replace=True, profile_name="z pos #1", control=False, color='blue'):
        if self.prop_plot_canvas[index] is None:
            self.prop_plot_canvas[index] = oasysgui.plotWindow(parent=None,
                                                               backend=None,
                                                               resetzoom=True,
                                                               autoScale=True,
                                                               logScale=True,
                                                               grid=True,
                                                               curveStyle=True,
                                                               colormap=False,
                                                               aspectRatio=False,
                                                               yInverted=False,
                                                               copy=True,
                                                               save=True,
                                                               print_=True,
                                                               control=control,
                                                               position=True,
                                                               roi=False,
                                                               mask=False,
                                                                fit=True)

            self.prop_plot_canvas[index].setDefaultPlotLines(True)
            self.prop_plot_canvas[index].setActiveCurveStyle(color="#00008B")
            self.prop_tab[index].layout().addWidget(self.prop_plot_canvas[index])

        title  = "Radial Intensity Profile"
        xtitle = "Radius [\u03bcm]"
        ytitle = "Intensity [A.U.]"

        self.prop_plot_canvas[index].setGraphTitle(title)
        self.prop_plot_canvas[index].setGraphXLabel(xtitle)
        self.prop_plot_canvas[index].setGraphYLabel(ytitle)

        rcParams['axes.formatter.useoffset']='False'

        self.prop_plot_canvas[index].addCurve(radius, profile_1D, profile_name, symbol='', color=color, xlabel=xtitle, ylabel=ytitle, replace=replace) #'+', '^', ','

        self.prop_plot_canvas[index].setInteractiveMode('zoom', color='orange')
        self.prop_plot_canvas[index].resetZoom()
        self.prop_plot_canvas[index].replot()

        self.prop_plot_canvas[index].setActiveCurve("Radial Intensity Profile")

    def plot_2D(self, index, dataX, dataY, data2D):
        origin = (dataX[0], dataY[0])
        scale = (dataX[1] - dataX[0], dataY[1] - dataY[0])

        colormap = {"name": "temperature", "normalization": "linear", "autoscale": True, "vmin": 0, "vmax": 0, "colors": 256}

        if self.prop_plot_canvas[index] is None:
            self.prop_plot_canvas[index] = Plot2D()

            self.prop_plot_canvas[index].resetZoom()
            self.prop_plot_canvas[index].setXAxisAutoScale(True)
            self.prop_plot_canvas[index].setYAxisAutoScale(True)
            self.prop_plot_canvas[index].setGraphGrid(False)
            self.prop_plot_canvas[index].setKeepDataAspectRatio(True)
            self.prop_plot_canvas[index].yAxisInvertedAction.setVisible(False)

            self.prop_plot_canvas[index].setXAxisLogarithmic(False)
            self.prop_plot_canvas[index].setYAxisLogarithmic(False)

            self.prop_plot_canvas[index].getMaskAction().setVisible(False)
            self.prop_plot_canvas[index].getRoiAction().setVisible(False)
            self.prop_plot_canvas[index].getColormapAction().setVisible(True)
            self.prop_plot_canvas[index].setKeepDataAspectRatio(False)

            self.prop_tab[index].layout().addWidget(self.prop_plot_canvas[index])
            
        self.prop_plot_canvas[index].clear()
        self.prop_plot_canvas[index].addImage(numpy.array(data2D),
                                              legend="rotated",
                                              scale=scale,
                                              origin=origin,
                                              colormap=colormap,
                                              replace=True)

        self.prop_plot_canvas[index].setActiveImage("rotated")
        self.prop_plot_canvas[index].setGraphXLabel("X' [\u03bcrad]")
        self.prop_plot_canvas[index].setGraphYLabel("Z' [\u03bcrad]")
        self.prop_plot_canvas[index].setGraphTitle("2D Divergence Profile")

add_widget_parameters_to_module(__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################

import numpy

import scipy.constants as codata
m2ev = codata.c * codata.h / codata.e

from AnyQt.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from oasys2.widget.widget import OWWidget, OWAction
from oasys2.widget import gui as oasysgui
from oasys2.widget.gui import ConfirmDialog, Styles
from oasys2.widget.util import congruence
from oasys2.widget.util.widget_objects import TriggerIn, TriggerOut
from oasys2.widget.util.exchange import DataExchangeObject

try:
    from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, ShadowPlot
except ImportError:
    pass

AUTOBINNING_FILE = "autobinning.dat"
FILTERS_FILE     = "filters.dat"
NONE             = numpy.nan

class EnergyBinning(object):
    def __init__(self,
                 energy_value = 0.0,
                 energy_value_to = None,
                 energy_step       = 0.0,
                 power_step        = None):
        self.energy_value    = energy_value
        self.energy_value_to = energy_value_to
        self.energy_step     = energy_step
        self.power_step      = power_step

    def __str__(self):
        return str(self.energy_value) + ", " + str(self.energy_value_to) + ", " + str(self.energy_step) + ", " + str(self.power_step)

class PowerLoopPoint(OWWidget):

    name = "Thermal Load Loop Point"
    description = "Tools: Thermal Loop Point"
    icon = "icons/cycle_power.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 6.0
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    class Inputs:
        trigger         = Input("Trigger", TriggerIn, default=True, auto_summary=False)
        energy_spectrum = Input("Energy Spectrum", DataExchangeObject, default=True, auto_summary=False)
        filters         = Input("Filters", DataExchangeObject, default=True, auto_summary=False)
    
    class Outputs:
        trigger = Output("Trigger", TriggerOut, default=True, auto_summary=False)
    
    want_main_area = 1

    _run_loop     = True
    _suspend_loop = False

    current_new_object       = 0
    number_of_new_objects    = 0
    total_current_new_object = 0
    total_new_objects        = Setting(0)

    energies              = Setting("")
    seed_increment        = Setting(1)
    autobinning           = Setting(1)
    auto_n_step           = Setting(1001)
    auto_perc_total_power = Setting(99)
    send_power_step       = Setting(0)

    load_file_mode        = Setting(1) # automatic
    skip_rows             = Setting(1)
    autobinning_file_name = Setting(AUTOBINNING_FILE)
    filters_file_name     = Setting(FILTERS_FILE)

    recovery_last_energy_step = -1
    recovery_file_name        = "<Partial Result before Crash>"

    current_energy_binning = -1
    current_energy_value   = NONE
    current_energy_step    = None
    current_power_step     = None

    energy_binnings  = None
    external_binning = False

    filters       = None
    spectrum_data = None

    #################################
    process_last = True
    #################################

    def __init__(self):
        self.runaction = OWAction("Start", self)
        self.runaction.triggered.connect(self.start_loop)
        self.addAction(self.runaction)

        self.runaction = OWAction("Stop", self)
        self.runaction.triggered.connect(self.stop_loop)
        self.addAction(self.runaction)

        runaction = OWAction("Suspend", self)
        runaction.triggered.connect(self.suspend_loop)
        self.addAction(runaction)

        runaction = OWAction("Restart", self)
        runaction.triggered.connect(self.restart_loop)
        self.addAction(runaction)

        runaction = OWAction(None, self)
        runaction.setSeparator(True)
        self.addAction(runaction)

        runaction = OWAction("Reload Spectrum and Filters", self)
        runaction.triggered.connect(self.read_spectrum_and_filters_file)
        self.addAction(runaction)

        runaction = OWAction("Reload Spectrum Only", self)
        runaction.triggered.connect(self.read_spectrum_file_only)
        self.addAction(runaction)

        self.setFixedWidth(1200)
        self.setFixedHeight(730)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=True, orientation="horizontal")

        self.start_button = gui.button(button_box, self, "Start", callback=self.start_loop)
        self.start_button.setFixedHeight(35)

        stop_button = gui.button(button_box, self, "Stop", callback=self.stop_loop)
        stop_button.setStyleSheet("color: red; font-weight: bold; font-style: italic;")
        stop_button.setFixedHeight(35)

        self.stop_button = stop_button

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=True, orientation="horizontal")

        suspend_button = gui.button(button_box, self, "Suspend", callback=self.suspend_loop)
        suspend_button.setStyleSheet("color: orange; font-weight: bold;")
        suspend_button.setFixedHeight(35)
        
        self.re_start_button = gui.button(button_box, self, "Restart", callback=self.restart_loop)
        self.re_start_button.setFixedHeight(35)
        self.re_start_button.setEnabled(False)

        tabs = oasysgui.tabWidget(self.controlArea)
        
        tab_loop     = oasysgui.createTabPage(tabs, "Loop")
        tab_recovery = oasysgui.createTabPage(tabs, "Advanced")

        left_box_1 = oasysgui.widgetBox(tab_loop, "", addSpace=False, orientation="vertical", width=385, height=560)

        oasysgui.lineEdit(left_box_1, self, "seed_increment", "Source Montecarlo Seed Increment", labelWidth=250, valueType=int, orientation="horizontal")

        gui.separator(left_box_1)

        gui.comboBox(left_box_1, self, "autobinning", label="Energy Binning",
                     items=["Manual", "Automatic (Constant Power)", "Automatic (Constant Energy)"], labelWidth=150,
                     callback=self.set_autobinning, sendSelectedValue=False, orientation="horizontal")

        self.autobinning_box_1 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=60)
        self.autobinning_box_2 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=155)

        gui.separator(self.autobinning_box_1)

        gui.button(self.autobinning_box_1, self, "Compute Bins", callback=self.calculate_energy_binnings, height=35)

        gui.separator(self.autobinning_box_1)

        header = oasysgui.widgetLabel(self.autobinning_box_1, "Energy From, Energy To, Energy Step [eV]")
        header.setStyleSheet("font-size: 10pt")

        # ----------------------------------------------

        oasysgui.lineEdit(self.autobinning_box_2, self, "auto_n_step", "Number of Steps", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(self.autobinning_box_2, self, "auto_perc_total_power", "% Total Power", labelWidth=250, valueType=float, orientation="horizontal")
        gui.comboBox(self.autobinning_box_2, self, "send_power_step", label="Send Power Step", items=["No", "Yes"], labelWidth=350, sendSelectedValue=False, orientation="horizontal")

        gui.separator(self.autobinning_box_2)

        button_box = oasysgui.widgetBox(self.autobinning_box_2, "", addSpace=False, orientation="horizontal")

        gui.button(button_box, self, "Reload Spectrum and Filters", callback=self.read_spectrum_and_filters_file, height=35)
        gui.button(button_box, self, "Reload Spectrum Only",        callback=self.read_spectrum_file_only, height=35)

        gui.separator(self.autobinning_box_2)

        header = oasysgui.widgetLabel(self.autobinning_box_2, "Energy Value [eV], Energy Step [eV], Power Step [W]")
        header.setStyleSheet("font-size: 10pt")

        def write_text(): self.energies = self.text_area.toPlainText()

        self.text_area = oasysgui.textArea(height=191, width=385, readOnly=False)
        self.text_area.setText(self.energies)
        self.text_area.setStyleSheet("background-color: white; font-family: Courier, monospace;")
        self.text_area.textChanged.connect(write_text)

        left_box_1.layout().addWidget(self.text_area)

        gui.separator(left_box_1)

        self.le_total_new_objects = oasysgui.lineEdit(left_box_1, self, "total_new_objects", "Total Energy Values", labelWidth=250, valueType=int, orientation="horizontal")
        self.le_total_new_objects.setReadOnly(True)
        self.le_total_new_objects.setStyleSheet(Styles.line_edit_read_only)

        self.le_number_of_new_objects = oasysgui.lineEdit(left_box_1, self, "number_of_new_objects", "Current Binning Energy Values", labelWidth=250, valueType=int, orientation="horizontal")
        self.le_number_of_new_objects.setReadOnly(True)
        self.le_number_of_new_objects.setStyleSheet(Styles.line_edit_read_only)

        gui.separator(left_box_1)

        le_total_current_value = oasysgui.lineEdit(left_box_1, self, "total_current_new_object", "Total New " + self.get_object_name(), labelWidth=250, valueType=int, orientation="horizontal")
        le_total_current_value.setReadOnly(True)
        le_total_current_value.setStyleSheet(Styles.line_edit_read_only)

        le_current_value = oasysgui.lineEdit(left_box_1, self, "current_new_object", "Current Binning New " + self.get_object_name(), labelWidth=250, valueType=int, orientation="horizontal")
        le_current_value.setReadOnly(True)
        le_current_value.setStyleSheet(Styles.line_edit_read_only)

        le_current_energy_value = oasysgui.lineEdit(left_box_1, self, "current_energy_value", "Current Energy Value", labelWidth=250, valueType=float, orientation="horizontal")
        le_current_energy_value.setReadOnly(True)
        le_current_energy_value.setStyleSheet(Styles.line_edit_read_only)

        # ----------------------------------------------

        left_box_2 = oasysgui.widgetBox(tab_recovery, "Input Files Management", addSpace=False, orientation="vertical", width=385, height=150)

        gui.comboBox(left_box_2, self, "load_file_mode", label="Input Files Mode",
                     items=["Manual", "Automatic"], labelWidth=150,
                     callback=self.set_load_file_mode, sendSelectedValue=False, orientation="horizontal")

        self.load_file_mode_box_1 = oasysgui.widgetBox(left_box_2, "", addSpace=False, orientation="vertical", height=100)
        self.load_file_mode_box_2 = oasysgui.widgetBox(left_box_2, "", addSpace=False, orientation="vertical", height=100)

        gui.separator(self.load_file_mode_box_2)
        oasysgui.lineEdit(self.load_file_mode_box_1, self, "skip_rows", "Skip Rows", labelWidth=260, valueType=int, orientation="horizontal")

        load_box = oasysgui.widgetBox(self.load_file_mode_box_1, "", addSpace=False, orientation="horizontal", height=30)
        self.le_autobinning_file_name = oasysgui.lineEdit(load_box, self, "autobinning_file_name", "Energy Spectrum", labelWidth=110, valueType=str, orientation="horizontal")
        gui.button(load_box, self, "...", callback=self.select_autobinning_file, width=25)

        load_box = oasysgui.widgetBox(self.load_file_mode_box_1, "", addSpace=False, orientation="horizontal", height=30)
        self.le_filters_file_name = oasysgui.lineEdit(load_box, self, "filters_file_name", "Filters File", labelWidth=110, valueType=str, orientation="horizontal")
        gui.button(load_box, self, "...", callback=self.select_filters_file, width=25)

        self.set_load_file_mode()

        left_box_2 = oasysgui.widgetBox(tab_recovery, "Crash Recovery", addSpace=False, orientation="vertical", width=385)

        oasysgui.lineEdit(left_box_2, self, "recovery_last_energy_step", "Last Energy Step #", labelWidth=260, valueType=int, orientation="horizontal")

        load_box = oasysgui.widgetBox(left_box_2, "", addSpace=False, orientation="horizontal", height=30)
        self.le_recovery_file_name = oasysgui.lineEdit(load_box, self, "recovery_file_name", "Extract from File", labelWidth=110, valueType=str, orientation="horizontal")
        gui.button(load_box, self, "...", callback=self.select_recovery_file, width=25)

        gui.button(left_box_2, self, "Restart from Crash", callback=self.restart_loop_from_crash, height=35)

        gui.rubber(self.controlArea)

        # ----------------------------------------------

        tabs = oasysgui.tabWidget(self.mainArea)
        tabs.setFixedHeight(self.height()-40)
        tabs.setFixedWidth(775)

        tab_plot = oasysgui.createTabPage(tabs, "Cumulated Power")
        tab_flux = oasysgui.createTabPage(tabs, "Spectral Flux")
        tab_fil = oasysgui.createTabPage(tabs, "Filter")

        self.cumulated_power_plot = oasysgui.plotWindow(tab_plot, position=True)
        self.cumulated_power_plot.setFixedHeight(self.height()-80)
        self.cumulated_power_plot.setFixedWidth(765)
        self.cumulated_power_plot.setGraphXLabel("Energy [eV]")
        self.cumulated_power_plot.setGraphYLabel("Cumulated Power [W]")
        self.cumulated_power_plot.setGraphTitle("Cumulated Power")

        self.spectral_flux_plot = oasysgui.plotWindow(tab_flux, position=True)
        self.spectral_flux_plot.setFixedHeight(self.height()-80)
        self.spectral_flux_plot.setFixedWidth(765)
        self.spectral_flux_plot.setGraphXLabel("Energy [eV]")
        self.spectral_flux_plot.setGraphYLabel("Flux [ph/s/.1%bw]")
        self.spectral_flux_plot.setGraphTitle("Spectral Flux")

        self.filter_plot = oasysgui.plotWindow(tab_fil, position=True)
        self.filter_plot.setFixedHeight(self.height()-80)
        self.filter_plot.setFixedWidth(765)
        self.filter_plot.setGraphXLabel("Energy [eV]")
        self.filter_plot.setGraphYLabel("Intensity Factor")
        self.filter_plot.setGraphTitle("Filter on Flux")

        self.set_autobinning()

    def set_autobinning(self):
        self.autobinning_box_1.setVisible(self.autobinning==0)
        self.autobinning_box_2.setVisible(self.autobinning==1 or self.autobinning==2)
        self.text_area.setReadOnly(self.autobinning>=1)
        self.text_area.setFixedHeight(185 if self.autobinning>=1 else 280)

    def set_load_file_mode(self):
        self.load_file_mode_box_1.setVisible(self.load_file_mode==0)
        self.load_file_mode_box_2.setVisible(self.load_file_mode==1)

    def select_autobinning_file(self):
        self.le_autobinning_file_name.setText(oasysgui.selectFileFromDialog(self, self.autobinning_file_name, "Select File", file_extension_filter="Text Files (*.txt *.dat)"))

    def select_recovery_file(self):
        self.le_recovery_file_name.setText(oasysgui.selectFileFromDialog(self, self.recovery_file_name, "Select File", file_extension_filter="HDF5 Files (*.hdf5)"))

        try:
            plot_file = ShadowPlot.PlotXYHdf5File(congruence.checkDir(self.recovery_file_name), mode="r")

            try:
                self.recovery_last_energy_step = plot_file.get_attribute(attribute_name="current_step", dataset_name="additional_data")
            except:
                raise ValueError("Last step non available in this file")

            plot_file.close()
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def select_filters_file(self):
        self.le_filters_file_name.setText(oasysgui.selectFileFromDialog(self, self.filters_file_name, "Select File", file_extension_filter="Text Files (*.txt *.dat)"))

    def read_spectrum_file(self, reset_filters=True):
        try:
            if self.load_file_mode == 1:
                autobinning_file_name = AUTOBINNING_FILE
                skip_rows = 1
            else:
                autobinning_file_name  = congruence.checkFile(self.autobinning_file_name)
                skip_rows              = congruence.checkPositiveNumber(self.skip_rows, "Skip Rows")

            if reset_filters:
                self.filters=None
                self.filter_plot.clear()

            data = numpy.loadtxt(autobinning_file_name, skiprows=skip_rows)

            calculated_data = DataExchangeObject(program_name="Shadow4", widget_name="PowerLoopPoint")
            calculated_data.add_content("spectrum_data", data)

            self.accept_energy_spectrum(calculated_data)
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def read_spectrum_file_only(self):
        self.read_spectrum_file(True)

    def read_spectrum_and_filters_file(self):
        try:
            if self.load_file_mode == 1:
                filter_file_name = FILTERS_FILE
                skip_rows = 1
            else:
                filter_file_name  = congruence.checkFile(self.filters_file_name)
                skip_rows         = congruence.checkPositiveNumber(self.skip_rows, "Skip Rows")

            data = numpy.loadtxt(filter_file_name, skiprows=skip_rows)

            calculated_data = DataExchangeObject(program_name="ShadowOui", widget_name="PowerLoopPoint")
            calculated_data.add_content("filters_data", data)

            self.accept_filters(calculated_data)
        except Exception as e:
            self.filters = None

            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

        self.read_spectrum_file(False)

    
    @Inputs.filters
    def accept_filters(self, exchange_data):
        if not exchange_data is None:
            try:
                try:
                    data = exchange_data.get_content("filters_data")
                    write_file = False
                except:
                    if not exchange_data.get_program_name() in ["XOPPY", "Shadow4-Thermal"]: raise ValueError("Only XOPPY or Shadow4-Thermal widgets are accepted")
                    if not exchange_data.get_widget_name() in ["XF1F2", "POWER", "XCRYSTAL", "MULTILAYER", "TOTAL_FILTER"]: raise ValueError("Only XF1F2, POWER, XCRYSTAL, MULTILAYER, Total Filter Calculator widgets are accepted")

                    if exchange_data.get_widget_name() == "TOTAL_FILTER":
                        data = exchange_data.get_content("total_filter")
                    elif exchange_data.get_widget_name() in ["XF1F2"]:
                        data = exchange_data.get_content("xoppy_data")
                    else:
                        input_data = exchange_data.get_content("xoppy_data")

                        if exchange_data.get_widget_name() == "POWER":
                            data = numpy.zeros((input_data.shape[0], 2))
                            data[:, 0] = input_data[:, 0]
                            data[:, 1] = input_data[:, -1]/input_data[:, 1]
                        else:
                            if exchange_data.get_widget_name() == "XCRYSTAL":
                                reflectivity_s = input_data[:, 5]
                                reflectivity_p = input_data[:, 6]
                            elif exchange_data.get_widget_name() == "MULTILAYER":
                                reflectivity_s = input_data[:, 1]
                                reflectivity_p = input_data[:, 2]
                            
                            reflectivity_s[numpy.where(numpy.isnan(reflectivity_s))] = 0
                            reflectivity_p[numpy.where(numpy.isnan(reflectivity_p))] = 0

                            data = numpy.zeros((input_data.shape[0], 2))
                            data[:, 0] = input_data[:, 0]
                            data[:, 1] = 0.5*(reflectivity_p + reflectivity_s)

                    write_file = True

                self.filters = data

                energies          = self.filters[:, 0]
                intensity_factors = self.filters[:, 1]

                if write_file:
                    file = open(FILTERS_FILE, "w")
                    file.write("Energy Filter")

                    for energy, intensity_factor in zip(energies, intensity_factors):
                        file.write("\n" + str(energy) + " " + str(intensity_factor))

                    file.flush()
                    file.close()

                    self.load_file_mode = 1 # back to automatic
                    self.set_load_file_mode()

                self.filter_plot.clear()
                self.filter_plot.addCurve(energies, intensity_factors, replace=True, legend="Intensity Factor")
                self.filter_plot.setGraphXLabel("Energy [eV]")
                self.filter_plot.setGraphYLabel("Intensity Factor")
                self.filter_plot.setGraphTitle("Filter on Flux")

                if self.autobinning==0:
                    if write_file: QMessageBox.information(self, "Info", "File filters.dat written on working directory, switch to Automatic binning to load it", QMessageBox.Ok)
                else:
                    if write_file: QMessageBox.information(self, "Info", "File filters.dat written on working directory", QMessageBox.Ok)

                if not self.spectrum_data is None:
                    calculated_data = DataExchangeObject(program_name="ShadowOui", widget_name="PowerLoopPoint")
                    calculated_data.add_content("spectrum_data", self.spectrum_data)

                    self.accept_energy_spectrum(calculated_data)

            except Exception as e:
                QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

                if self.IS_DEVELOP: raise e
    
    @Inputs.energy_spectrum
    def accept_energy_spectrum(self, exchange_data):
        if not exchange_data is None:
            try:
                try:
                    data = exchange_data.get_content("spectrum_data")
                    write_file = False
                except:
                    try:    data = exchange_data.get_content("srw_data")
                    except: data = exchange_data.get_content("xoppy_data")
                    write_file = True

                self.spectrum_data = data.copy()

                energies                              = data[:, 0]
                flux_through_finite_aperture          = data[:, 1]
                flux_through_finite_aperture_filtered = flux_through_finite_aperture.copy()

                if not self.filters is None:
                    flux_through_finite_aperture_filtered *= numpy.interp(energies, self.filters[:, 0], self.filters[:, 1])
                    use_filters = True
                else:
                    use_filters = False

                if write_file:
                    file = open(AUTOBINNING_FILE, "w")
                    file.write("Energy Flux")

                    for energy, flux in zip(energies, flux_through_finite_aperture):
                        file.write("\n" + str(energy) + " " + str(flux))

                    file.flush()
                    file.close()

                    self.load_file_mode = 1 # back to automatic
                    self.set_load_file_mode()

                if self.autobinning==0:
                    if write_file: QMessageBox.information(self, "Info", "File autobinning.dat written on working directory, switch to Automatic binning to load it", QMessageBox.Ok)
                else:
                    if write_file: QMessageBox.information(self, "Info", "File autobinning.dat written on working directory", QMessageBox.Ok)

                    congruence.checkStrictlyPositiveNumber(self.auto_n_step, "(Auto) % Number of Steps")
                    congruence.checkStrictlyPositiveNumber(self.auto_perc_total_power, "(Auto) % Total Power")

                    energy_step = energies[1] - energies[0]

                    # last energy do not contribute to the total (the approximated integral of the power is out of the range)

                    power           = flux_through_finite_aperture * (1e3 * energy_step * codata.e)
                    cumulated_power = numpy.cumsum(power)
                    total_power     = cumulated_power[-1]

                    if use_filters:
                        power_filtered           = flux_through_finite_aperture_filtered * (1e3 * energy_step * codata.e)
                        cumulated_power_filtered = numpy.cumsum(power_filtered)
                        total_power_filtered     = cumulated_power_filtered[-1]

                    self.text_area.clear()

                    self.cumulated_power_plot.clear()
                    self.spectral_flux_plot.clear()

                    if use_filters:
                        total_power_binning                  = total_power_filtered
                        cumulated_power_binning              = cumulated_power_filtered
                        flux_through_finite_aperture_binning = flux_through_finite_aperture_filtered
                    else:
                        total_power_binning                  = total_power
                        cumulated_power_binning              = cumulated_power
                        flux_through_finite_aperture_binning = flux_through_finite_aperture

                    good = numpy.where(cumulated_power_binning <= self.auto_perc_total_power*0.01*total_power_binning)

                    energies                             = energies[good]
                    cumulated_power                      = cumulated_power[good]
                    cumulated_power_binning              = cumulated_power_binning[good]
                    flux_through_finite_aperture         = flux_through_finite_aperture[good]
                    flux_through_finite_aperture_binning = flux_through_finite_aperture_binning[good]

                    if use_filters:
                        cumulated_power_filtered = cumulated_power_filtered[good]
                        flux_through_finite_aperture_filtered = flux_through_finite_aperture_filtered[good]

                    if self.autobinning==1: # constant power
                        interpolated_cumulated_power = numpy.linspace(start=numpy.min(cumulated_power_binning), stop=numpy.max(cumulated_power_binning), num=self.auto_n_step+1)
                        interpolated_energies        = numpy.interp(interpolated_cumulated_power, cumulated_power_binning, energies)
                        energy_steps = numpy.ediff1d(interpolated_energies)

                        interpolated_energies        = interpolated_energies[:-1]
                        interpolated_cumulated_power = interpolated_cumulated_power[:-1]

                        power_steps  = numpy.ones(self.auto_n_step)*cumulated_power_binning[-1]/self.auto_n_step

                    elif self.autobinning==2: # constant energy
                        minimum_energy = energies[0]
                        maximum_energy = energies[-1]
                        energy_step = (maximum_energy-minimum_energy)/self.auto_n_step

                        interpolated_energies        = numpy.arange(minimum_energy, maximum_energy, energy_step)
                        interpolated_cumulated_power = numpy.interp(interpolated_energies, energies, cumulated_power_binning)

                        energy_steps = numpy.ones(self.auto_n_step)*energy_step
                        power_steps  = numpy.ediff1d(numpy.append(numpy.zeros(1), interpolated_cumulated_power))

                    flux_steps = numpy.interp(interpolated_energies, energies, flux_through_finite_aperture_binning)

                    self.energy_binnings = []
                    self.total_new_objects = 0

                    if not use_filters:
                        self.cumulated_power_plot.addCurve(energies, cumulated_power, replace=True, legend="Cumulated Power")
                    else:
                        self.cumulated_power_plot.addCurve(energies, cumulated_power, replace=True, linestyle="--", legend="Cumulated Power")
                        self.cumulated_power_plot.addCurve(energies, cumulated_power_filtered, replace=False, legend="Cumulated Power (Filtered)", color="#006400")
                    self.cumulated_power_plot.setGraphXLabel("Energy [eV]")
                    self.cumulated_power_plot.setGraphYLabel("Cumulated " + ("" if self.filters is None else " (Filtered)") + " Power" )
                    self.cumulated_power_plot.setGraphTitle("Total Power: " + str(round(power_steps.sum(), 2)) + " W")

                    if not use_filters:
                        self.spectral_flux_plot.addCurve(energies, flux_through_finite_aperture, replace=True, legend="Spectral Flux")
                    else:
                        self.spectral_flux_plot.addCurve(energies, flux_through_finite_aperture, replace=True, linestyle="--", legend="Spectral Flux")
                        self.spectral_flux_plot.addCurve(energies, flux_through_finite_aperture_filtered, replace=False, legend="Spectral Flux (Filtered)", color="#006400")
                    self.spectral_flux_plot.setGraphXLabel("Energy [eV]")
                    self.spectral_flux_plot.setGraphYLabel("Flux [ph/s/.1%bw]")
                    self.spectral_flux_plot.setGraphTitle("Spectral Flux" + ("" if use_filters else " (Filtered)"))

                    self.cumulated_power_plot.addCurve(interpolated_energies, interpolated_cumulated_power, replace=False, legend="Energy Binning", color="red", linestyle=" ", symbol="+")
                    self.spectral_flux_plot.addCurve(interpolated_energies, flux_steps, replace=False, legend="Energy Binning", color="red", linestyle=" ", symbol="+")

                    self.cumulated_power_plot.getLegendsDockWidget().setVisible(True)
                    self.spectral_flux_plot.getLegendsDockWidget().setVisible(True)

                    text = ""
                    for energy_value, energy_step, power_step in zip(interpolated_energies, energy_steps, power_steps):
                        energy_binning = EnergyBinning(energy_value=round(energy_value, 3),
                                                       energy_step=round(energy_step, 3),
                                                       power_step=round(power_step, 4))

                        text += str(round(energy_value, 3)) + ", " + \
                                str(round(energy_step, 3))  + ", " + \
                                str(round(power_step, 4)) + "\n"

                        self.energy_binnings.append(energy_binning)
                        self.total_new_objects += 1

                    self.text_area.setText(text)

                    self.external_binning = True
            except Exception as e:
                QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

                if self.IS_DEVELOP: raise e
        else:
            self.energy_binnings = None
            self.total_new_objects = 0
            self.external_binning = False
            self.text_area.setText("")

    def calculate_energy_binnings(self):
        if not self.external_binning:
            self.total_new_objects = 0

            rows = self.energies.split("\n")
            for row in rows:
                data = row.split(",")
                if len(data) == 3:
                    if self.energy_binnings is None: self.energy_binnings = []

                    energy_from = float(data[0].strip())
                    energy_to   = float(data[1].strip())
                    energy_step = float(data[2].strip())

                    energy_binning = EnergyBinning(energy_value=energy_from,
                                                   energy_value_to=energy_to,
                                                   energy_step=energy_step)
                    self.energy_binnings.append(energy_binning)

                    self.total_new_objects += int((energy_to - energy_from) / energy_step)

    def calculate_number_of_new_objects(self):
        if len(self.energy_binnings) > 0:
            if self.external_binning:
                self.number_of_new_objects = 1
            else:
                energy_binning = self.energy_binnings[self.current_energy_binning]

                self.number_of_new_objects = int((energy_binning.energy_value_to - energy_binning.energy_value) / energy_binning.energy_step)
        else:
            self.number_of_new_objects = 0

    def reset_values(self):
        self.current_new_object = 0
        self.total_current_new_object = 0
        self.current_energy_value = NONE
        self.current_energy_step = None
        self.current_energy_binning = -1
        self.current_power_step = None

        if not self.external_binning: self.energy_binnings = None

    def start_loop(self):
        try:
            if self.spectrum_data is None: raise ValueError("Initialize Loop point from Spectrum Data (Reload or from Widget)")

            self.calculate_energy_binnings()

            self.current_new_object = 1
            self.total_current_new_object = 1
            self.current_energy_binning = 0
            self.current_energy_value             = round(self.energy_binnings[0].energy_value, 8)
            self.current_energy_step              = round(self.energy_binnings[0].energy_step, 8)
            self.current_power_step               = None if self.energy_binnings[0].power_step is None else (None if self.send_power_step==0 else round(self.energy_binnings[0].power_step, 8))
            self.calculate_number_of_new_objects()

            self.start_button.setEnabled(False)
            self.text_area.setEnabled(False)
            self.setStatusMessage("Running " + self.get_object_name() + " " + str(self.total_current_new_object) + " of " + str(self.total_new_objects))
            self.Outputs.trigger.send(TriggerOut(new_object=True,
                                                 additional_parameters={"energy_value"   : self.current_energy_value,
                                                                        "energy_step"    : self.current_energy_step,
                                                                        "power_step"     : -1 if self.current_power_step is None else self.current_power_step,
                                                                        "seed_increment" : self.seed_increment,
                                                                        "current_step"   : self.total_current_new_object,
                                                                        "total_steps"    : self.total_new_objects,
                                                                        "start_event"    : True}))
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass

    def stop_loop(self):
        try:
            if ConfirmDialog.confirmed(parent=self, message="Confirm Interruption of the Loop?"):
                self._run_loop = False
                self.reset_values()
                self.setStatusMessage("Interrupted by user")
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass

    def suspend_loop(self):
        try:
            if ConfirmDialog.confirmed(parent=self, message="Confirm Suspension of the Loop?"):
                self._run_loop = False
                self._suspend_loop = True
                self.stop_button.setEnabled(False)
                self.re_start_button.setEnabled(True)
                self.setStatusMessage("Suspended by user")
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass

    def restart_loop_from_crash(self):
        try:
            if self.energy_binnings is None: raise Exception("Reload Spectrum (and Filters, if present) before restarting the loop")

            self.current_energy_binning   = self.recovery_last_energy_step + 1
            self.total_current_new_object = self.recovery_last_energy_step + 1

            if self.current_energy_binning < len(self.energy_binnings):
                energy_binning = self.energy_binnings[self.current_energy_binning]

                self.current_new_object = 1
                self.calculate_number_of_new_objects()
                self.current_energy_value = round(energy_binning.energy_value, 8)
                self.current_power_step = None if energy_binning.power_step is None else (None if self.send_power_step == 0 else round(energy_binning.power_step, 8))

                self.setStatusMessage("Running " + self.get_object_name() + " " + str(self.total_current_new_object) + " of " + str(self.total_new_objects))
                self.start_button.setEnabled(False)
                self.text_area.setEnabled(False)
                self.Outputs.trigger.send(TriggerOut(new_object=True,
                                                additional_parameters={"energy_value": self.current_energy_value,
                                                                       "energy_step": energy_binning.energy_step,
                                                                       "power_step": -1 if self.current_power_step is None else self.current_power_step,
                                                                       "current_step": self.total_current_new_object,
                                                                       "total_steps": self.total_new_objects,
                                                                       "seed_increment": self.seed_increment,
                                                                       "start_event": False}))
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def restart_loop(self):
        try:
            self._run_loop = True
            self._suspend_loop = False
            self.stop_button.setEnabled(True)
            self.re_start_button.setEnabled(False)
            self.pass_trigger(TriggerIn(new_object=True))
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass

    def get_object_name(self):
        return "Shadow4 Beam"

    @Inputs.trigger
    def pass_trigger(self, trigger):
        if self._run_loop:
            if trigger:
                if trigger.interrupt:
                    self.reset_values()
                    self.start_button.setEnabled(True)
                    self.text_area.setEnabled(True)
                    self.setStatusMessage("")
                    self.Outputs.trigger.send(TriggerOut(new_object=False))
                elif trigger.new_object:
                    if self.energy_binnings is None: self.calculate_energy_binnings()

                    if self.current_energy_binning == -1:
                        QMessageBox.critical(self, "Error", "Power Loop has to be started properly: press the button Start", QMessageBox.Ok)
                        return

                    if self.current_energy_binning < len(self.energy_binnings):
                        energy_binning = self.energy_binnings[self.current_energy_binning]

                        self.total_current_new_object += 1

                        if self.current_new_object < self.number_of_new_objects:
                            if self.current_energy_value == NONE:
                                self.current_new_object = 1
                                self.calculate_number_of_new_objects()
                                self.current_energy_value = round(energy_binning.energy_value, 8)
                            else:
                                self.current_new_object += 1
                                self.current_energy_value = round(self.current_energy_value + energy_binning.energy_step, 8)

                            self.current_power_step = None if energy_binning.power_step is None else (None if self.send_power_step==0 else round(energy_binning.power_step, 8))

                            self.setStatusMessage("Running " + self.get_object_name() + " " + str(self.total_current_new_object) + " of " + str(self.total_new_objects))
                            self.start_button.setEnabled(False)
                            self.text_area.setEnabled(False)
                            self.Outputs.trigger.send(TriggerOut(new_object=True,
                                                                 additional_parameters={"energy_value"   : self.current_energy_value,
                                                                                        "energy_step"    : energy_binning.energy_step,
                                                                                        "power_step"     : -1 if self.current_power_step is None else self.current_power_step,
                                                                                        "current_step"   : self.total_current_new_object,
                                                                                        "total_steps"    : self.total_new_objects,
                                                                                        "seed_increment" : self.seed_increment,
                                                                                        "start_event"    : False}))
                        else:
                            self.current_energy_binning += 1

                            if self.current_energy_binning < len(self.energy_binnings):
                                energy_binning = self.energy_binnings[self.current_energy_binning]

                                self.current_new_object = 1
                                self.calculate_number_of_new_objects()
                                self.current_energy_value = round(energy_binning.energy_value, 8)
                                self.current_power_step = None if energy_binning.power_step is None else (None if self.send_power_step==0 else round(energy_binning.power_step, 8))

                                self.setStatusMessage("Running " + self.get_object_name() + " " + str(self.total_current_new_object) + " of " + str(self.total_new_objects))
                                self.start_button.setEnabled(False)
                                self.text_area.setEnabled(False)
                                self.Outputs.trigger.send(TriggerOut(new_object=True,
                                                                     additional_parameters={"energy_value"   : self.current_energy_value,
                                                                                            "energy_step"    : energy_binning.energy_step,
                                                                                            "power_step"     : -1 if self.current_power_step is None else self.current_power_step,
                                                                                            "current_step"   : self.total_current_new_object,
                                                                                            "total_steps"    : self.total_new_objects,
                                                                                            "seed_increment" : self.seed_increment,
                                                                                            "start_event"    : False}))
                            else:
                                self.reset_values()
                                self.start_button.setEnabled(True)
                                self.text_area.setEnabled(True)
                                self.setStatusMessage("")
                                self.Outputs.trigger.send(TriggerOut(new_object=False))
                    else:
                        self.reset_values()
                        self.start_button.setEnabled(True)
                        self.text_area.setEnabled(True)
                        self.setStatusMessage("")
                        self.Outputs.trigger.send(TriggerOut(new_object=False))
        else:
            if not self._suspend_loop:
                self.reset_values()
                self.start_button.setEnabled(True)
                self.text_area.setEnabled(True)

            self.Outputs.trigger.send(TriggerOut(new_object=False))
            self.setStatusMessage("")
            self._suspend_loop = False
            self._run_loop = True


add_widget_parameters_to_module(__name__)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################


from orangewidget.settings import Setting
from orangewidget.widget import Input

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.util.exchange import DataExchangeObject

from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

try:
    from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, ShadowPlot
except ImportError:
    pass

import scipy.constants as codata

from orangecontrib.shadow4_advanced.widgets.gui.power_plot_xy import AbstractPowerPlotXY, PowerPlotXYWidget

class PowerPlotXYBM(AbstractPowerPlotXY):

    name = "Power Density - BM"
    description = "Display Data Tools: Power Plot XY - BM"
    icon = "icons/bm_plot_xy_power.png"
    priority = 7

    class Inputs:
        shadow_data = AbstractPowerPlotXY.Inputs.shadow_data
        energy_spectrum = Input("Energy Spectrum", DataExchangeObject, default=True, auto_summary=False)

    nbins_interpolation = Setting(500)

    initial_flux = None
    initial_energy = None

    def __init__(self):
        super().__init__()

    def _set_additional_boxes(self, tab_gen):
        interpolation_box = oasysgui.widgetBox(tab_gen, "Interpolation", addSpace=True, orientation="vertical", height=65)

        oasysgui.lineEdit(interpolation_box, self, "nbins_interpolation", "Number of Bins for energy interpolation", labelWidth=250, valueType=int, orientation="horizontal")

    def plot_cumulated_data(self):
        if not self.input_data is None: self.plot_results()

    def replace_fig(self, shadow_beam, var_x, var_y, xrange, yrange, nbins_h, nbins_v, nolost):
        if self.initial_flux is None: raise ValueError("Energy Spectrum is not set")

        if self.plot_canvas is None:
            self.plot_canvas = PowerPlotXYWidget()
            self.image_box.layout().addWidget(self.plot_canvas)
        else:
            self.plot_canvas.clear()

        try:
            ticket = self.plot_canvas.plot_power_density_BM(shadow_beam, self.initial_energy, self.initial_flux, self.nbins_interpolation,
                                                            var_x, var_y, nbins_h, nbins_v, xrange, yrange, nolost,
                                                            show_image=self.view_type==1)

            self.plotted_ticket = ticket
            self.plotted_ticket_original = self.plotted_ticket.copy()

        except Exception as e:
            if not self.IS_DEVELOP:
                raise Exception("Data not plottable: " + str(e))
            else:
                raise e

    def _check_other_fields(self):
        self.nbins_interpolation = congruence.checkStrictlyPositiveNumber(self.nbins_interpolation, "Number of Bins for energy interpolation")

    #########################################################
    # I/O

    @Inputs.shadow_data
    def set_shadow_data(self, input_data):
        super(PowerPlotXYBM, self).set_shadow_data(input_data)

    def _analyze_input_data(self, input_data):
        self.input_data = input_data.duplicate()

        if self.input_data.scanning_data and self.input_data.scanning_data.has_additional_parameter("is_footprint"): # from footprint reader
            if self.input_data.scanning_data.get_additional_parameter("is_footprint"):
                self.cb_rays.setEnabled(False)
                self.rays = 0  # transmitted, absorbed doesn't make sense since is precalculated by footprint object
            else:
                self.cb_rays.setEnabled(True)

        return True

    def _can_be_plotted(self, input_data):
        if super(PowerPlotXYBM, self)._can_be_plotted(input_data): return not self.initial_flux is None
        else: return False



    @Inputs.energy_spectrum
    def set_flux(self, exchange_data):
        if not exchange_data is None:
            if exchange_data.get_program_name() == "XOPPY" and exchange_data.get_widget_name() == "BM":
                    if exchange_data.get_content("is_log_plot") == 1:
                        raise Exception("Logarithmic X scale of Xoppy Energy distribution not supported")
                    elif exchange_data.get_content("calculation_type") == 0 and (exchange_data.get_content("psi") == 0 or exchange_data.get_content("psi") == 2):
                        index_flux = 5
                    else:
                        raise Exception("Xoppy result is not an Flux vs Energy distribution integrated in a rectangular space")
            elif exchange_data.get_program_name() == "XOPPY" and (exchange_data.get_widget_name() == "WS" or exchange_data.get_widget_name() == "XWIGGLER"):
                index_flux = 1
            else:
                raise Exception("Exchange data are not from a XOPPY BM widget")

            spectrum = exchange_data.get_content("xoppy_data")
            self.initial_flux = spectrum[:, index_flux]
            self.initial_energy = spectrum[:, 0]

            initial_energy_step = self.initial_energy[1] - self.initial_energy[0]

            self.total_initial_power = self.initial_flux.sum() * 1e3 * codata.e * initial_energy_step

            print("Total Initial Power from XOPPY", self.total_initial_power)

            if not self.input_data is None:
                if ShadowCongruence.check_empty_data(self.input_data):
                    if ShadowCongruence.check_good_beam(self.input_data.beam):
                        self.plot_results()

        else:
            self.initial_flux = None
            self.initial_energy = None

add_widget_parameters_to_module(__name__)